{-
Part of "Haski the Robot"
examination project for adv. funct. progr. 2004

    Copyright (C) 2004  Lars Otten

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

-}
module Haski (haskiRun) where

import Graphics.UI.WX
import Graphics.UI.WXCore hiding (run,Timer,size)
import MapLexer
import MapParser
import HaskiLexer
import HaskiParser
import System
import Directory
import List

---------------------------------------------------------
-- some global constants

-- sets the height and width of the working area
-- !! CAREFUL !!
-- Changing this will require updating maps as well !
areaWidth = 20
areaHeight = 15

-- sets the size (height=width) of a single field, new graphics are
-- needed when changing this (put in directory /data/<fieldsize> )
fieldsize = 24
-- used for internal computations
fsize = fieldsize+1

-- sets the default base speed of the robots [ms]
speed = 50

---------------------------------------------------------
-- image definitions

-- be careful: images-paths are dependant on fieldsize-value
robotUp = bitmap ("data/" ++ show fieldsize ++ "/robotUp.png")
robotDo = bitmap ("data/" ++ show fieldsize ++ "/robotDo.png")
robotLe = bitmap ("data/" ++ show fieldsize ++ "/robotLe.png")
robotRi = bitmap ("data/" ++ show fieldsize ++ "/robotRi.png")

ob1 = bitmap ("data/" ++ show fieldsize ++ "/1.png")
ob2 = bitmap ("data/" ++ show fieldsize ++ "/2.png")
ob3 = bitmap ("data/" ++ show fieldsize ++ "/3.png")
ob4 = bitmap ("data/" ++ show fieldsize ++ "/4.png")
ob5 = bitmap ("data/" ++ show fieldsize ++ "/5.png")
ob6 = bitmap ("data/" ++ show fieldsize ++ "/6.png")
ob7 = bitmap ("data/" ++ show fieldsize ++ "/7.png")
ob8 = bitmap ("data/" ++ show fieldsize ++ "/8.png")
ob9 = bitmap ("data/" ++ show fieldsize ++ "/9.png")
ob10 = bitmap ("data/" ++ show fieldsize ++ "/10.png")

background = bitmap "data/background.jpg"

---------------------------------------------------------
-- data type definitions

type Wall = (Int,Int)
type Item = (Int,Int)
type Position = Int

-- keeps the state of the area the robots are working on
data Area = Area{
 -- items in the area, format (position,quantity)
 objects :: [Item] ,
 -- walls in the area, format (a,b) if wall is between field a & b
 walls :: [Wall],
 -- the map-file which was parsed for the area
 mapfile :: FilePath
 }

-- keeps track of a robot's state
data State = State{
 numItems :: Int,	-- no. of items the robot is currently carrying
 facing :: Facing,	-- direction robot is currently facing
 pos :: Position,	-- current position
 prog :: Program,	-- the main program
 funcs :: [Function],	-- all functions returned by the lexer/parser
 file :: FilePath	-- the hs-file which was parsed for that robot
 }

data Facing
 = FUp
 | FLeft
 | FDown
 | FRight
   deriving (Eq,Show)

{-
-- to be found in the parsing modules
-- HaskiParser and HaskiLexer

data Sensor = ...
data Program = ...
-}

---------------------------------------------------------
-- commands available to the user for the embedded language

{-
-- are not needed anymore after introducing the
-- parsing of external files

-- for combining two programs into one
(&>) :: Program -> Program -> Program
p &> q = Seq p q

-- normal commands
go_forward :: Program
turn_right :: Program
take_item :: Program
drop_item :: Program
do_nothing :: Program

go_forward = Forward
turn_right = Turn
take_item = Take
drop_item = Drop
do_nothing = Idle

-- sensor commands
front_free :: Sensor
right_free :: Sensor
left_free :: Sensor
field_has_item :: Sensor
is_carrying :: Sensor
facing_up :: Sensor
facing_right :: Sensor
facing_down :: Sensor
facing_left :: Sensor

front_free = Front_free
right_free = Right_free
left_free = Left_free
field_has_item = Field_has_item
is_carrying = Is_carrying
facing_up = Facing_up
facing_right = Facing_right
facing_down = Facing_down
facing_left = Facing_left
-}


---------------------------------------------------------
-- main program
-- reads and parses files, invokes the GUI

-- gives the path to the maps
mapPath = "maps/"
mapFileEnding = ".map"

-- give the path to the robot-progams
progPath = "programs/"
progFileEnding = ".hs"

-- checks if the specified files exist and starts the GUI
-- first parameter is the mapname (without .map), second parameter
-- is the list of program-names (without .hs)
haskiRun :: String -> [String] -> IO ()
haskiRun mapfile' progs' =
 do let mapfile = (mapPath ++ mapfile' ++ mapFileEnding)
    mapOK <- doesFileExist mapfile
    if not mapOK then putStrLn ("\nCouldn't find map-file \""
                                 ++ mapfile ++ "\", aborting!")
                      >> exitFailure
                 else return ()
    let progs = map (\x -> progPath ++ x ++ progFileEnding) progs'
    (progsOK,missingFile) <- checkProgs progs
    if not progsOK then putStrLn ("\nCouldn't find Haski-codefile \""
                                 ++ missingFile ++ "\", aborting!")
                      >> exitFailure
                 else return ()
    start $ run mapfile progs


-- checks for all passed codefile-names if they exist
-- on error returns (False, <name of first failed file> )
checkProgs :: [String] -> IO (Bool,String)
checkProgs [] = return (True,"")
checkProgs (f:fs) =
 do b <- doesFileExist f
    if not b then return (False,f)
             else checkProgs fs

---------------------------------------------------------
-- main window

-- the number of lines to be kept in the log
logNoLines = 100

windowWidth  = areaWidth*fsize + 1
windowHeight = areaHeight*fsize + 1

-- runs the GUI and executes the program(s)
run :: FilePath -> [FilePath] -> IO ()
run mapf progs =
 do 
    win <- frameFixed [text := "Haski the Robot", resizeable := False]
    logField <- textCtrl win [ enabled := True, wrap := WrapNone,
                               outerSize := sz (fsize*areaWidth) 100 ]
    textCtrlMakeLogActiveTarget logField
    (area',robots') <- loadMapAndHs mapf progs
    area <- varCreate area'
    robots <- varCreate robots'
    robotfiles <- varCreate progs
    pan <- panel win [ bgcolor := white,
                        layout := space windowWidth windowHeight,
                        on paint := drawAll area robots ]
    time <- timer pan [interval := speed*5,
                       enabled := False ]
    -- buttons and speed-slider
    pauseB <- button win [text := "Start"]
    quitB <- button win [text := "Exit",
                        on command := close win]
    singleStepB <- button win [text := "Single Step",
                              on command := cropLog logField >> doStep area robots >> repaint pan ]
    resetB <- button win [text := "Reset", on command :=
                          set logField [ text := "" ] >>
                          disableT time pauseB singleStepB >>
                          resetAll area robots robotfiles >> repaint pan ]
    speedChange <- hslider win False 1 20 [selection := 15,
                                           outerSize := sz 100 20]
    -- the status line
    status <- statusField   [text := "Welcome to Haski !"]
    -- building the menu
    menuFile <- menuPane [ text := "&File" ]
    menuFileLoadmap <- menuItem menuFile [ text := "Load &map\tCtrl+M",
                         on command := disableT time pauseB singleStepB >>
                         loadNewMap win logField area robots robotfiles >> repaint pan ]
    menuFileLoadhs <- menuItem menuFile [ text := "Load &haski\tCtrl+H",
                         on command := disableT time pauseB singleStepB >>
                         loadNewHs win logField area robots robotfiles >> repaint pan ]
    menuLine menuFile
    menuFileQuit <- menuQuit menuFile [ help := "Quit Haski", on command := close win ]
    menuHelp <- menuPane [ text := "&Help" ]
    menuHelpAbout <- menuItem menuHelp [ text := "&About\tF1", on command :=
                      infoDialog win "About Haski the Robot"
                      "Version 1.0, December 2004\n\nDeveloped by Lars Otten\n  <haski@lotten.net>\nhttp://lotten.net/uni/haski/" ]
    -- setting some properties
    set pauseB [on command := toggleT time pauseB singleStepB]
    set pan [on (charKey 'p') := toggleT time pauseB singleStepB,
             on (charKey 's') := doStep area robots >> repaint pan,
             on (charKey 'q') := close win]
    set time [ on command := cropLog logField >>
                             doStep area robots >>= \b ->
                             (if b then return ()
                                   else disableT time pauseB singleStepB)
                             >> repaint pan ]

    set speedChange [ on command := changeSpeed speedChange time ]
    set win [statusBar := [status], menubar := [menuFile,menuHelp], layout := column 5 
              [ alignCenter $ widget pan,
                alignCenter $ widget logField,
                alignCenter $ row 5
                 [widget pauseB, widget singleStepB,
                  alignCenter$ label "Speed:",
                  alignCenter $ widget speedChange,
                  widget resetB,
                  widget quitB] ] ]


-- reads all haski-codefiles and parses them
parseAllProgs :: [FilePath] -> IO [[Function]]
parseAllProgs [] = return []
parseAllProgs (p:ps) =
  do content <- readFile p
     remaining <- parseAllProgs ps
     case haskiParser content of
       Ok fs -> return (fs:remaining)
       Failed e -> do logMessage ("\nERROR in " ++ p ++ " : " ++ e ++ "\n  -> Using empty program instead")
                      return ([Def "main" End]:remaining)

-- assigns a list of functions to the respective state,
-- also saving the filename of the program
assignProgFuncs :: [State] -> [[Function]] -> [FilePath] -> IO [State]
assignProgFuncs states fcts files =
 mapM assignIt (zip3 states fcts files)
  where assignIt (s,f,fil) = do q <- findProg "main" fil f
                                return s{funcs=f,prog=q,file=fil}

-- reloads and reparses the complete map and robot-instructions
-- according to the filenames inside the data structure
resetAll :: Var Area -> Var [State] -> Var [FilePath] -> IO ()
resetAll areaV rsV fsV =
 do Area{mapfile=mapf} <- varGet areaV
    rs <- varGet rsV
    files <- varGet fsV
    (area,states) <- loadMapAndHs mapf files
    varSet areaV area
    varSet rsV states
  where getFileNameFromState State{file=f} = f

-- loads and parses the map file and all hs-files for the robots
loadMapAndHs :: FilePath -> [FilePath] -> IO (Area,[State])
loadMapAndHs mapf hsfiles = 
 do functs <- parseAllProgs hsfiles
    (area,robots') <- parseMapFile mapf
    robots <- assignProgFuncs robots' functs hsfiles
    return (area,robots)


-- lets the user choose a new mapfile, loads and parses it
loadNewMap :: Frame () -> TextCtrl () -> Var Area -> Var [State] -> Var [FilePath] -> IO ()
loadNewMap win logF areaV rsV fsV =
 do file' <- fileOpenDialog win False True "Load mapfile" [("Map file",["*"++mapFileEnding])] mapPath ""
    case file' of
      Nothing -> return ()
      Just f  -> do set logF [ text := "" ]
                    z@Area{mapfile=mapf} <- varGet areaV
                    varSet areaV z{mapfile=f}
                    resetAll areaV rsV fsV

-- lets the user choose a new hs-file, loads and parses it
-- This will remove all robots except the first one 
loadNewHs :: Frame () -> TextCtrl () -> Var Area -> Var [State] -> Var [FilePath] -> IO ()
loadNewHs win logF areaV rsV fsV =
 do file' <- fileOpenDialog win False True "Load mapfile" [("Haski file",["*"++progFileEnding])] progPath ""
    case file' of
      Nothing -> return ()
      Just f  -> do set logF [ text := "" ]
                    (r:rs) <- varGet rsV -- there's always at least one robot !
                    varSet rsV [r{file=f}]
                    varSet fsV [f]
                    resetAll areaV rsV fsV

-- crops the contents of the loggging text-field
-- Otherwise, for long programs, at some point no additional entries
-- can be made.
cropLog :: TextCtrl () -> IO ()
cropLog logF =
 do tex <- get logF text
    set logF [ text := unlines $ crop $ lines tex ]
     where crop xs = case length' xs > 200 of
                      True  -> drop (length' xs - logNoLines) xs
                      False -> xs
           length' xs = length xs


-- changes the timer speed
changeSpeed :: Slider () -> Timer -> IO ()
changeSpeed s t =
 do val <- get s selection
    set t [interval:=(21-val)*speed]

-- switches the timer on and off and modifys the respective buttons
toggleT timer pausBut stepBut
 = do set pausBut [text :~ switchText]
      set stepBut [enabled :~ not]
      set timer [enabled :~ not]
       where switchText s = if s=="Pause" then "Continue" else "Pause"

-- disables the timer
disableT timer pausBut stepBut
 = do set pausBut [text := "Start" ]
      set stepBut [enabled := True ]
      set timer [enabled := False]

-- executes one step for all robots
doStep :: Var Area -> Var [State] -> IO Bool
doStep area states =
 do area' <- varGet area
    states' <- varGet states
    (newArea,newStates,logs) <- processAll area' states' [] []
    let messages = emptyCheck (addNumbers 1 logs)
    mapM_ logMessage $ messages
    varSet area newArea
    varSet states newStates
    return (messages/=[])
     where addNumbers _ [] = []
           addNumbers n ("":xs) = addNumbers (n+1) xs
           addNumbers n (x:xs) = ("Robot " ++ show n ++ ": " ++ x)
                                  :(addNumbers (n+1) xs)
           emptyCheck [] = []
           emptyCheck xs = " - - - - - - - - - - - - -":xs

-- helper function for doStep
processAll :: Area -> [State] -> [State] -> [String] -> IO (Area,[State],[String])
processAll a (x:xs) ys ms =
 do (a',x',mes) <- process a x
    processAll a' xs (ys++[x']) (mes:ms)
processAll a [] ys ms = return (a,ys,reverse ms)


-- takes one command of the robot and executes it
-- updates the area and the state accordingly
-- requires extensive, ugly pattern-matching
process :: Area -> State -> IO (Area,State,String)

process a
        s@State{prog=End} =
 return (a,s,"")

--------- taking care of Seq-constructs --------
process a
        s@State{prog=Seq End q} =
 process a s{prog=q}

process a
        s@State{prog=Seq p q} =
 do (a',s'@State{prog=p'},logm) <- process a s{prog=p}
    return (a',s'{prog=Seq p' q},logm)

--------- Idle ---------
process a
        s@State{prog=Idle} =
 return (a,s{prog=End}, "Idling")

--------- Forward ---------
process a@Area{objects=o,walls=w}
        s@State{prog=Forward,pos=p,facing=f} =
 return (a,s{prog=End,pos=p'},mes p p')
  where p' = executeMove w p f
        mes p p' = if p/=p' then "Moving forward"
                            else "Movement blocked"

--------- Turn ---------
process a@Area{objects=o}
        s@State{prog=Turn,facing=f} =
 return (a,s{prog=End,facing=executeTurn f},"Turning right")

--------- Take ---------
process a@Area{objects=o}
        s@State{prog=Take,pos=p,numItems=n} =
 do let (o',n') = executeTake o p n
    return (a{objects=o'},s{prog=End,numItems=n'},mes n n')
     where mes n n' = if n/=n' then "Taking item, now carrying " ++ show n'
                               else "Failed taking item"

--------- Drop ---------
process a@Area{objects=o}
        s@State{prog=Drop,pos=p,numItems=n} =
 do let (o',n') = executeDrop o p n
    return (a{objects=o'},s{prog=End,numItems=n'},mes n n')
     where mes n n' = if n/=n' then "Dropping item, now carrying " ++ show n'
                               else "Failed dropping item"

--------- IfThen ---------
process a
        s@State{prog=IfThen co pr} =
 do let cond = evaluateCondition co a s
    if cond
     then return (a,s{prog=pr},"IfThen-condition satisfied, executing")
     else return (a,s{prog=End},"IfThen-condition NOT satisfied, skipping")

--------- If ---------
process a
        s@State{prog=IfThenElse co pr1 pr2} =
 do let cond = evaluateCondition co a s
    if cond
     then return (a,s{prog=pr1},"IfThenElse-condition satisfied, executing first option")
     else return (a,s{prog=pr2},"IfThenElse-condition NOT satisfied, executing second option")

--------- While ---------
process a
        s@State{prog=While co pr} =
 do let cond = evaluateCondition co a s
    if cond
     then return (a,s{prog=Seq pr (While co pr)},"While-condition satisfied, executing")
     else return (a,s{prog=End},"While-condition NOT satisfied, leaving loop")

--------- Call ---------
process a
        s@State{prog=Call p,funcs=fs,file=f} =
 findProg p f fs >>= \q -> process a s{prog=q}


---------------------------------------------------------
-- Sensor-related functions

-- lots of pattern matching again
evaluateCondition :: Sensor -> Area -> State -> Bool

--------- AND ---------
evaluateCondition (x :& y) a s =
 evaluateCondition x a s && evaluateCondition y a s

--------- OR ---------
evaluateCondition (x :| y) a s =
 evaluateCondition x a s || evaluateCondition y a s

--------- Not ---------
evaluateCondition (Not x) a s =
 not $ evaluateCondition x a s

--------- Facing_up ---------
evaluateCondition Facing_up
                  _
                  State{facing=f} =
 if f==FUp then True else False

--------- Facing_Right ---------
evaluateCondition Facing_right
                  _
                  State{facing=f} =
 if f==FRight then True else False

--------- Facing_down ---------
evaluateCondition Facing_down
                  _
                  State{facing=f} =
 if f==FDown then True else False

--------- Facing_left ---------
evaluateCondition Facing_left
                  _
                  State{facing=f} =
 if f==FLeft then True else False

--------- Is_carrying ---------
evaluateCondition Is_carrying
                  _
                  State{numItems=n} =
 if n>0 then True else False

--------- Is_carrying ---------
evaluateCondition Field_has_item
                  Area{objects=os}
                  State{pos=p} =
 if elem p (map fst os) then True else False

--------- Front_free ---------
evaluateCondition Front_free
                  Area{walls=ws}
                  State{pos=p,facing=f} =
 not ( (elem (p,p+d) ws) || (elem (p+d,p) ws) || (f==FUp && y==0) || (f==FLeft && x==0)
       || (f==FDown && y==areaHeight-1) || (f==FRight && x==areaWidth-1) )
   where (x,y) = getXY p
         d = case f of
              FUp    -> -areaWidth
              FRight -> 1
              FDown  -> areaWidth
              FLeft  -> -1

--------- Right_free ---------
evaluateCondition Right_free
                  Area{walls=ws}
                  State{pos=p,facing=f} =
 not ( (elem (p,p+d) ws) || (elem (p+d,p) ws) || (f==FLeft && y==0) || (f==FDown && x==0)
       || (f==FRight && y==areaHeight-1) || (f==FUp && x==areaWidth-1) )
   where (x,y) = getXY p
         d = case f of
              FUp    -> 1
              FRight -> areaWidth
              FDown  -> -1
              FLeft  -> -areaWidth

--------- Left_free ---------
evaluateCondition Left_free
                  Area{walls=ws}
                  State{pos=p,facing=f} =
 not ( (elem (p,p+d) ws) || (elem (p+d,p) ws) || (f==FRight && y==0) || (f==FUp && x==0)
       || (f==FLeft && y==areaHeight-1) || (f==FDown && x==areaWidth-1) )
   where (x,y) = getXY p
         d = case f of
              FUp    -> -1
              FRight -> -areaWidth
              FDown  -> 1
              FLeft  -> areaWidth


---------------------------------------------------------
-- little reusable helper functions for the processing

executeTurn :: Facing -> Facing
executeTurn f = case f of
                    FUp    -> FRight
                    FRight -> FDown
                    FDown  -> FLeft
                    FLeft  -> FUp


executeMove :: [Wall] -> Position -> Facing -> Position
executeMove ws p f =
 if move_possible ws p (diff f)
  then p+(diff f)
  else p
    where diff f = case f of
                    FUp    -> -areaWidth
                    FRight -> 1
                    FDown  -> areaWidth
                    FLeft  -> -1
          move_possible ws p d = not (
                                     p+d < 0 ||
                                     p+d > areaWidth*areaHeight -1 ||
                                     elem (p,p+d) ws ||
                                     elem (p+d,p) ws ||
                                     not (isAdjacentField p (p+d))
                                     )
            where isAdjacentField p q = let (p1,p2) = getXY p
                                            (q1,q2) = getXY q
                                        in p1==q1 || q2==p2 

executeTake :: [Item] -> Position -> Int -> ([Item],Int)
executeTake os p n =
 if elem p (map fst os)
  then (takeOne p os,n+1)
  else (os,n)
    where takeOne p ((q,m):os) = if p/=q then (q,m):(takeOne p os)
                                         else if m>1 then (q,m-1):os
                                                     else takeOne p os
          takeOne _ [] = []

executeDrop :: [Item] -> Position -> Int -> ([Item],Int)
executeDrop os p n =
 if n>0
 then
  if elem p (map fst os)
   then (addOne p os,n-1)
   else ((p,1):os,n-1)
 else (os,n)
     where addOne p ((q,m):os) = if p/=q then (q,m):(addOne p os)
                                         else (q,m+1):os
           addOne _ [] = []

---------------------------------------------------------
-- findProg
-- searches a robots functions and returns the respective program
-- if function can't be found it outputs an error message and returns
-- the empty program

findProg :: String -> FilePath -> [Function] -> IO Program
findProg s f [] = logMessage ( "\nERROR in " ++ f ++ " : Runtime error: function \""
                             ++ s ++ "\" not found\n ->  Skipping call")
                >> return Idle
findProg s f ((Def s' p):xs) =
 if s==s' then return p
          else findProg s f xs


---------------------------------------------------------
-- drawing functions

drawAll :: Var Area -> Var [State] -> DC () -> Rect -> IO ()
drawAll a s dc r =
 do bgPos <- centerBG background
    drawBitmap dc background bgPos True []
    drawGrid dc
    drawBorder dc
    a' <- varGet a
    s' <- varGet s
    drawArea a' dc
    drawStates s' dc
  where centerBG pic =
         do Size w h <- get pic size
            let x = div (fsize * areaWidth - w) 2
            let y = div (fsize * areaHeight - h) 2
            return $ pt x y

drawGrid :: DC () -> IO ()
drawGrid dc =
 do drawHorizGrid dc fsize (fsize*areaHeight) (fsize*areaWidth)
    drawVertGrid dc fsize (fsize*areaWidth) (fsize*areaHeight)

drawHorizGrid :: DC () -> Int -> Int -> Int -> IO ()
drawHorizGrid dc height border right =
 if height < border
  then drawLineG dc (pt 0 height, pt right height)
       >> drawHorizGrid dc (height+fsize) border right
  else return ()

drawVertGrid :: DC () -> Int -> Int -> Int -> IO ()
drawVertGrid dc width border bottom =
 if width < border
  then drawLineG dc (pt width 0, pt width bottom)
       >> drawVertGrid dc (width+fsize) border bottom
  else return ()


drawBorder :: DC () -> IO ()
drawBorder dc =
 do drawLine dc (pt 0 0, pt rightborder 0 )
    drawLine dc (pt 0 0, pt 0 lowerborder )
    drawLine dc (pt 0 lowerborder, pt (rightborder+1) lowerborder )
    drawLine dc (pt rightborder 0, pt rightborder (lowerborder+1) )
   where lowerborder = fsize*areaHeight
         rightborder = fsize*areaWidth

drawArea :: Area -> DC () -> IO ()
drawArea Area{objects=o,walls=w} dc =
 do mapM_ (drawObj dc) o
    mapM_ (drawWall dc) w

drawObj :: DC () -> (Int,Int) -> IO ()
drawObj dc (p,n) = drawBitmap dc pic (pos p) True []
  where pic = case n of
               1 -> ob1
               2 -> ob2
               3 -> ob3
               4 -> ob4
               5 -> ob5
               6 -> ob6
               7 -> ob7
               8 -> ob8
               9 -> ob9
               _ -> ob10
        pos p = pt (x* fsize + 1) (y* fsize + 1)
          where (x,y) = getXY p

drawWall :: DC () -> (Int,Int) -> IO()
drawWall dc (a,b) = case (relation xa ya xb yb) of
                     "right" -> drawLine dc (pt
                                              ( (xa+1)*fsize )
                                              ( (ya)*fsize ) ,
                                             pt
                                              ( (xa+1)*fsize )
                                              ( (ya+1)*fsize +1 ) )
                     "left"  -> drawLine dc (pt
                                              ( (xa)*fsize )
                                              ( (ya)*fsize ) ,
                                             pt
                                              ( (xa)*fsize )
                                              ( (ya+1)*fsize +1 ) )

                     "above" -> drawLine dc (pt
                                              ( (xa)*fsize )
                                              ( (ya)*fsize ) ,
                                             pt
                                              ( (xa+1)*fsize +1 )
                                              ( (ya)*fsize ) )

                     "below" -> drawLine dc (pt
                                              ( (xa)*fsize )
                                              ( (ya+1)*fsize ) ,
                                             pt
                                              ( (xa+1)*fsize +1 )
                                              ( (ya+1)*fsize ) )
  where (xa,ya) = getXY a
        (xb,yb) = getXY b
        relation xa ya xb yb = if xa==xb-1
                                then "right"
                                else if xa==xb+1
                                 then "left"
                                 else if ya==yb-1
                                  then "below"
                                  else "above"

drawStates :: [State] -> DC () -> IO ()
drawStates states dc = mapM_ (drawState dc) states

drawState :: DC () -> State -> IO ()
drawState dc State{pos=p,facing=face} =
 drawBitmap dc pic (pos p) True []
   where pic = case face of
                FUp    -> robotUp
                FRight -> robotRi
                FDown  -> robotDo
                FLeft  -> robotLe
         pos p = pt (x* fsize + 1) (y* fsize + 1)
          where (x,y) = getXY p


drawLine dc (x1,x2) = line dc x1 x2 [penColor := red]
drawLineG dc (x1,x2) = line dc x1 x2 [penColor := lightgrey]

-- for a given field in single-Int numbering,
-- this returns the (x,y)-value of the field
getXY :: Int -> (Int,Int)
getXY n = ( (mod n areaWidth),
            (div n areaWidth) )

-- for a given field in X/Y-coordinate
-- this returns the single-Int coordinate of the field
getPos :: (Int,Int) -> Int
getPos (x,y) = x + (y * areaWidth)

---------------------------------------------------------
-- reading and parsing of map

parseMapFile :: FilePath -> IO (Area,[State])
parseMapFile file =
 do file' <- readFile file
    case mapParser file' of
      Failed e -> do logMessage ("\nERROR in " ++ file ++ " : " ++ e ++ "\n  -> Using empty map instead")
                     return ( Area{walls=[], objects=[], mapfile=file} ,
                              [State{pos=0,numItems=0,facing=FRight,prog=End,funcs=[],file=""}] )
      Ok info  -> transformMapData info file

transformMapData :: [MapData] -> FilePath -> IO (Area,[State])
transformMapData dat file =
  rs dat >>= \rss -> return ( Area{walls=ws dat,objects=os dat,mapfile=file} , rss )
  where ws dat = nub $ filter checkWall [ w | Wall w <- dat ]
        os dat = [ (p,n) | Item (p,n) <- dat, n>0, p<300 ]
        rs dat = case tuples==[] of -- having no robot at all is not allowed
                  False -> return $ map tupleToState tuples
                  True  -> logMessage "\nNo robot-definition found in mapfile, correcting..."
                           >> return [tupleToState (0,0,0)]
         where tuples = [ r | Robot r <- dat ]


tupleToState :: (Int,Int,Int) -> State
tupleToState (p,n,f) = State{pos=checkPos
                            ,numItems=checkNum
                            ,facing=face
                            ,prog=End,funcs=[],file=""}
  where face = case f of
                1 -> FRight
                2 -> FDown
                3 -> FLeft
                _ -> FUp
        checkNum = case n<0 of
                    True  -> 0
                    False -> n
        checkPos = case p<300 of
                    True  -> p
                    False -> 299

-- checks if a given wall is possible after all, i.e. whether the
-- two fields are adjacent
checkWall :: (Int,Int) -> Bool
checkWall (a,b) = (ya==yb && (xa==xb+1 || xb==xa+1))
                   || (xa==xb && (ya==yb+1 || yb==ya+1))
  where (xa,ya) = getXY a
        (xb,yb) = getXY b


---------------------------------------------------------
-- End of File
